@@ -8,23 +8,32 @@
 #define _FCITX5_FRONTEND_WAYLANDIM_WAYLANDIMSERVERBASE_H_
 
 #include <string>
-#include <wayland-client-core.h>
 #include <xkbcommon/xkbcommon.h>
-#include "fcitx-utils/event.h"
 #include "fcitx-utils/misc.h"
-#include "fcitx/focusgroup.h"
 #include "display.h"
 #include "waylandim.h"
+#include "wl_seat.h"
 
 namespace fcitx {
 
 class WaylandIMServerBase {
 public:
     WaylandIMServerBase(wl_display *display, FocusGroup *group,
-                        const std::string &name, WaylandIMModule *waylandim);
+                        std::string name, WaylandIMModule *waylandim);
     virtual ~WaylandIMServerBase() = default;
 
-    void deferredFlush();
+    auto *parent() { return parent_; }
+    auto *display() { return display_; }
+
+    std::optional<std::string> mayCommitAsText(const Key &key,
+                                               uint32_t state) const;
+
+    int32_t repeatRate(
+        const std::shared_ptr<wayland::WlSeat> &seat,
+        const std::optional<std::tuple<int32_t, int32_t>> &defaultValue) const;
+    int32_t repeatDelay(
+        const std::shared_ptr<wayland::WlSeat> &seat,
+        const std::optional<std::tuple<int32_t, int32_t>> &defaultValue) const;
 
 protected:
     FocusGroup *group_;
@@ -32,13 +41,16 @@ protected:
     WaylandIMModule *parent_;
     wayland::Display *display_;
 
-    std::unique_ptr<EventSource> deferEvent_;
-
     UniqueCPtr<struct xkb_context, xkb_context_unref> context_;
     UniqueCPtr<struct xkb_keymap, xkb_keymap_unref> keymap_;
     UniqueCPtr<struct xkb_state, xkb_state_unref> state_;
 
     KeyStates modifiers_;
+
+private:
+    std::optional<std::tuple<int32_t, int32_t>> repeatInfo(
+        const std::shared_ptr<wayland::WlSeat> &seat,
+        const std::optional<std::tuple<int32_t, int32_t>> &defaultValue) const;
 };
 
 } // namespace fcitx
