@@ -11,6 +11,7 @@
 #include <thread>
 #include <fcitx-utils/event.h>
 #include <fcitx-utils/eventdispatcher.h>
+#include <fcitx-utils/trackableobject.h>
 #include <xcb/xcb.h>
 #include "xcb_public.h"
 
@@ -18,7 +19,7 @@ namespace fcitx {
 
 class XCBConnection;
 
-class XCBEventReader {
+class XCBEventReader : public TrackableObject<XCBEventReader> {
 public:
     XCBEventReader(XCBConnection *conn);
     ~XCBEventReader();
@@ -37,13 +38,13 @@ private:
     void run();
     bool onIOEvent(IOEventFlags flags);
     XCBConnection *conn_;
-    EventDispatcher dispatcherToMain_;
+    EventDispatcher &dispatcherToMain_;
     EventDispatcher dispatcherToWorker_;
     bool hadError_ = false;
     std::unique_ptr<EventSource> deferEvent_;
     std::unique_ptr<EventSource> wakeEvent_;
+    std::unique_ptr<EventSource> postEvent_;
     std::unique_ptr<std::thread> thread_;
-    std::unique_ptr<EventLoop> event_;
     std::mutex mutex_;
     std::list<UniqueCPtr<xcb_generic_event_t>> events_;
 };
