@@ -6,12 +6,16 @@
  */
 
 #include "xcbui.h"
+#include <cairo.h>
 #include <xcb/randr.h>
 #include <xcb/xcb_aux.h>
 #include <xcb/xinerama.h>
-#include "fcitx-utils/stringutils.h"
+#include <xcb/xproto.h>
+#include "fcitx-utils/endian_p.h"
+#include "fcitx-utils/misc.h"
 #include "xcbinputwindow.h"
 #include "xcbtraywindow.h"
+#include "xcbwindow.h"
 
 namespace fcitx::classicui {
 
@@ -99,13 +103,13 @@ XCBFontOption forcedDpi(xcb_connection_t *conn, xcb_screen_t *screen) {
     parse(resources, "Xft.rgba:\t", [&option](const std::string &value) {
         if (value == "none") {
             option.rgba = XCBRGBA::NoRGBA;
-        } else if (value == "hintnone") {
+        } else if (value == "rgb") {
             option.rgba = XCBRGBA::RGB;
-        } else if (value == "hintmedium") {
+        } else if (value == "bgr") {
             option.rgba = XCBRGBA::BGR;
-        } else if (value == "hintslight") {
+        } else if (value == "vrgb") {
             option.rgba = XCBRGBA::VRGB;
-        } else if (value == "hintslight") {
+        } else if (value == "vbgr") {
             option.rgba = XCBRGBA::VBGR;
         }
         // default
@@ -176,23 +180,32 @@ void XCBFontOption::setupPangoContext(PangoContext *context) const {
 
 XCBUI::XCBUI(ClassicUI *parent, const std::string &name, xcb_connection_t *conn,
              int defaultScreen)
-    : parent_(parent), name_(name), conn_(conn), defaultScreen_(defaultScreen) {
-    ewmh_ = parent_->xcb()->call<IXCBModule::ewmh>(name_);
+    : UIInterface("x11:" + name), parent_(parent), displayName_(name),
+      conn_(conn), defaultScreen_(defaultScreen) {
+    ewmh_ = parent_->xcb()->call<IXCBModule::ewmh>(displayName_);
     inputWindow_ = std::make_unique<XCBInputWindow>(this);
     trayWindow_ = std::make_unique<XCBTrayWindow>(this);
 
     compMgrAtomString_ = "_NET_WM_CM_S" + std::to_string(defaultScreen_);
     compMgrAtom_ = parent_->xcb()->call<IXCBModule::atom>(
-        name_, compMgrAtomString_, false);
+        displayName_, compMgrAtomString_, false);
 
     auto xsettingsSelectionString =
         "_XSETTINGS_S" + std::to_string(defaultScreen_);
     managerAtom_ =
-        parent_->xcb()->call<IXCBModule::atom>(name_, "MANAGER", false);
+        parent_->xcb()->call<IXCBModule::atom>(displayName_, "MANAGER", false);
     xsettingsSelectionAtom_ = parent_->xcb()->call<IXCBModule::atom>(
-        name_, xsettingsSelectionString, false);
+        displayName_, xsettingsSelectionString, false);
     xsettingsAtom_ = parent_->xcb()->call<IXCBModule::atom>(
-        name_, "_XSETTINGS_SETTINGS", false);
+        displayName_, "_XSETTINGS_SETTINGS", false);
+
+    initScreenEvent_ = parent_->instance()->eventLoop().addTimeEvent(
+        CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 10000, 0,
+        [this](EventSourceTime *, uint64_t) {
+            initScreen();
+            return true;
+        });
+    initScreenEvent_->setEnabled(false);
 
     eventHandlers_.emplace_back(parent_->xcb()->call<IXCBModule::addSelection>(
         name, compMgrAtomString_,
@@ -231,6 +244,13 @@ XCBUI::XCBUI(ClassicUI *parent, const std::string &name, xcb_connection_t *conn,
                         property->window == xsettingsWindow_) {
                         readXSettings();
                     }
+
+                    auto *screen = xcb_aux_get_screen(conn_, defaultScreen_);
+                    if (property->window == screen->root &&
+                        property->atom == XCB_ATOM_RESOURCE_MANAGER) {
+                        fontOption_ = forcedDpi(conn_, screen);
+                    }
+
                     break;
                 }
                 case XCB_CONFIGURE_NOTIFY: {
@@ -238,24 +258,33 @@ XCBUI::XCBUI(ClassicUI *parent, const std::string &name, xcb_connection_t *conn,
                         reinterpret_cast<xcb_configure_notify_event_t *>(event);
                     auto *screen = xcb_aux_get_screen(conn_, defaultScreen_);
                     if (configure->window == screen->root) {
-                        initScreen();
+                        scheduleUpdateScreen();
                     }
                     break;
                 }
                 }
-                if (multiScreen_ == MultiScreenExtension::Randr &&
-                    xrandrFirstEvent_ == XCB_RANDR_NOTIFY) {
-                    auto *randr =
-                        reinterpret_cast<xcb_randr_notify_event_t *>(event);
-                    if (randr->subCode == XCB_RANDR_NOTIFY_CRTC_CHANGE) {
-                        initScreen();
+                if (multiScreen_ == MultiScreenExtension::Randr) {
+                    if (response_type ==
+                        xrandrFirstEvent_ + XCB_RANDR_SCREEN_CHANGE_NOTIFY) {
+                        scheduleUpdateScreen();
+                    } else if (response_type ==
+                               xrandrFirstEvent_ + XCB_RANDR_NOTIFY) {
+
+                        auto *randr =
+                            reinterpret_cast<xcb_randr_notify_event_t *>(event);
+                        if (randr->subCode == XCB_RANDR_NOTIFY_CRTC_CHANGE ||
+                            randr->subCode == XCB_RANDR_NOTIFY_OUTPUT_CHANGE) {
+                            scheduleUpdateScreen();
+                        }
                     }
                 }
                 return false;
             }));
 
     xcb_screen_t *screen = xcb_aux_get_screen(conn_, defaultScreen_);
-    addEventMaskToWindow(conn_, screen->root, XCB_EVENT_MASK_STRUCTURE_NOTIFY);
+    addEventMaskToWindow(conn_, screen->root,
+                         XCB_EVENT_MASK_STRUCTURE_NOTIFY |
+                             XCB_EVENT_MASK_PROPERTY_CHANGE);
     root_ = screen->root;
     fontOption_ = forcedDpi(conn_, screen);
     CLASSICUI_DEBUG() << "Xft.dpi: " << fontOption_.dpi;
@@ -265,17 +294,26 @@ XCBUI::XCBUI(ClassicUI *parent, const std::string &name, xcb_connection_t *conn,
     refreshManager();
 }
 
-XCBUI::~XCBUI() {}
+XCBUI::~XCBUI() {
+    inputWindow_.reset();
+    trayWindow_.reset();
+    device_.reset();
+}
 
 void XCBUI::initScreen() {
     auto *screen = xcb_aux_get_screen(conn_, defaultScreen_);
     int newScreenCount = xcb_setup_roots_length(xcb_get_setup(conn_));
-    if (newScreenCount == 1) {
+    if (multiScreen_ == MultiScreenExtension::EXTNone && newScreenCount == 1) {
         const xcb_query_extension_reply_t *reply =
             xcb_get_extension_data(conn_, &xcb_randr_id);
         if (reply && reply->present) {
             multiScreen_ = MultiScreenExtension::Randr;
             xrandrFirstEvent_ = reply->first_event;
+            xcb_randr_select_input(conn_, screen->root,
+                                   XCB_RANDR_NOTIFY_MASK_SCREEN_CHANGE |
+                                       XCB_RANDR_NOTIFY_MASK_OUTPUT_CHANGE |
+                                       XCB_RANDR_NOTIFY_MASK_CRTC_CHANGE |
+                                       XCB_RANDR_NOTIFY_MASK_OUTPUT_PROPERTY);
         } else {
             const xcb_query_extension_reply_t *reply =
                 xcb_get_extension_data(conn_, &xcb_xinerama_id);
@@ -355,16 +393,17 @@ void XCBUI::initScreen() {
                         if (crtc) {
                             Rect rect;
                             rect.setPosition(crtc->x, crtc->y);
-                            int dpiX = 25.4 * crtc->width / output->mm_width;
-                            int dpiY = 25.4 * crtc->height / output->mm_height;
+                            auto outputWidth = output->mm_width;
+                            auto outputHeight = output->mm_height;
                             if (crtc->rotation ==
                                     XCB_RANDR_ROTATION_ROTATE_90 ||
                                 crtc->rotation ==
                                     XCB_RANDR_ROTATION_ROTATE_270) {
-                                rect.setSize(crtc->height, crtc->width);
-                            } else {
-                                rect.setSize(crtc->width, crtc->height);
+                                std::swap(outputWidth, outputHeight);
                             }
+                            int dpiX = 25.4 * crtc->width / outputWidth;
+                            int dpiY = 25.4 * crtc->height / outputHeight;
+                            rect.setSize(crtc->width, crtc->height);
                             rects_.emplace_back(rect, std::min(dpiX, dpiY));
                             if (maxDpi_ < rects_.back().second) {
                                 maxDpi_ = rects_.back().second;
@@ -392,7 +431,7 @@ void XCBUI::initScreen() {
                 auto y = info->y_org;
                 auto w = info->width;
                 auto h = info->height;
-                rects_.emplace_back(Rect(x, y, x + w - 1, y + h - 1), -1);
+                rects_.emplace_back(Rect().setPosition(x, y).setSize(w, h), -1);
             }
         }
     }
@@ -402,8 +441,12 @@ void XCBUI::initScreen() {
             Rect(0, 0, screen->width_in_pixels, screen->height_in_pixels), -1);
     }
 
+    screenDpi_ =
+        25.4 * screen->height_in_pixels / screen->height_in_millimeters;
+
     CLASSICUI_DEBUG() << "Screen rects are: " << rects_
-                      << " Primary DPI: " << primaryDpi_;
+                      << " Primary DPI: " << primaryDpi_
+                      << " XScreen DPI: " << screenDpi_;
 }
 
 void XCBUI::refreshCompositeManager() {
@@ -415,15 +458,24 @@ void XCBUI::refreshCompositeManager() {
     }
 
     xcb_screen_t *screen = xcb_aux_get_screen(conn_, defaultScreen_);
+    if (needFreeColorMap_) {
+        xcb_free_colormap(conn_, colorMap_);
+    }
+
     if (compMgrWindow_) {
         addEventMaskToWindow(conn_, compMgrWindow_,
                              XCB_EVENT_MASK_STRUCTURE_NOTIFY);
         colorMap_ = xcb_generate_id(conn_);
         xcb_create_colormap(conn_, XCB_COLORMAP_ALLOC_NONE, colorMap_,
                             screen->root, visualId());
+        needFreeColorMap_ = true;
     } else {
         colorMap_ = screen->default_colormap;
+        needFreeColorMap_ = false;
     }
+    CLASSICUI_DEBUG() << "Refresh color map: " << colorMap_
+                      << " vid: " << visualId()
+                      << " CompMgr: " << compMgrWindow_;
     inputWindow_->createWindow(visualId());
     // mainWindow_->createWindow();
 }
@@ -442,7 +494,6 @@ void XCBUI::refreshManager() {
                                  XCB_EVENT_MASK_PROPERTY_CHANGE);
     }
     xcb_ungrab_server(conn_);
-    xcb_flush(conn_);
 
     readXSettings();
 }
@@ -478,19 +529,11 @@ void XCBUI::readXSettings() {
         }
     } while (more);
     xcb_ungrab_server(conn_);
-    xcb_flush(conn_);
 
     if (error || data.empty()) {
         return;
     }
-    enum { BYTE_ORDER_MSB_FIRST = 1, BYTE_ORDER_LSB_FIRST = 0 };
-    const uint16_t endian = 1;
-    uint8_t byteOrder = 0;
-    if (*reinterpret_cast<const char *>(&endian)) {
-        byteOrder = BYTE_ORDER_LSB_FIRST;
-    } else {
-        byteOrder = BYTE_ORDER_MSB_FIRST;
-    }
+    auto byteOrder = hostByteOrder();
     if (data[0] != BYTE_ORDER_LSB_FIRST && data[0] != BYTE_ORDER_MSB_FIRST) {
         return;
     }
@@ -555,6 +598,9 @@ void XCBUI::readXSettings() {
     if (!readCard32(&nSettings)) {
         return;
     }
+    if (static_cast<uint64_t>(nSettings) * 8 + 8 > data.size()) {
+        return;
+    }
     for (uint32_t i = 0; i < nSettings; i++) {
         // 1      SETTING_TYPE  type
         // 1                    unused
@@ -579,7 +625,7 @@ void XCBUI::readXSettings() {
         if (!readCard16(&nameLen)) {
             return;
         }
-#define XSETTINGS_PAD(n, m) ((n + m - 1) & (~(m - 1)))
+#define XSETTINGS_PAD(n, m) (((n) + (m) - 1) & (~((m) - 1)))
         uint32_t namePad = XSETTINGS_PAD(nameLen, 4);
         if (std::distance(iter, data.cend()) < namePad) {
             return;
@@ -647,7 +693,10 @@ void XCBUI::updateCursor(InputContext *inputContext) {
     inputWindow_->updatePosition(inputContext);
 }
 
-void XCBUI::updateCurrentInputMethod(InputContext *) { trayWindow_->update(); }
+void XCBUI::updateCurrentInputMethod(InputContext *inputContext) {
+    FCITX_UNUSED(inputContext);
+    trayWindow_->update();
+}
 
 int XCBUI::dpiByPosition(int x, int y) {
     int shortestDistance = INT_MAX;
@@ -664,21 +713,28 @@ int XCBUI::dpiByPosition(int x, int y) {
 }
 
 int XCBUI::scaledDPI(int dpi) {
-    if (!*parent_->config().perScreenDPI) {
-        // CLASSICUI_DEBUG() << "Use font dpi: " << fontOption_.dpi;
-        return fontOption_.dpi;
+    if (!*parent_->config().perScreenDPI ||
+        parent_->xcb()->call<IXCBModule::isXWayland>(displayName_)) {
+        // CLASSICUI_DEBUG() << "Use font option dpi: " << fontOption_.dpi;
+        if (fontOption_.dpi > 0) {
+            return fontOption_.dpi;
+        }
+        if (screenDpi_ >= 96) {
+            // Nowadays their should not be tiny dpi screen I assume.
+            // In ancient days, there used to be invalid DPI value that make
+            // font extremely tiny.
+            return screenDpi_;
+        }
+        return -1;
     }
     if (dpi < 0) {
         return fontOption_.dpi;
     }
 
     double targetDPI;
-    if (fontOption_.dpi < 0) {
-        targetDPI = dpi;
-    } else {
-        auto baseDPI = primaryDpi_ > 0 ? primaryDpi_ : maxDpi_;
-        targetDPI = (static_cast<double>(dpi) / baseDPI) * fontOption_.dpi;
-    }
+    auto baseScreenDPI = primaryDpi_ > 0 ? primaryDpi_ : maxDpi_;
+    auto baseDPI = fontOption_.dpi > 0 ? fontOption_.dpi : screenDpi_;
+    targetDPI = (static_cast<double>(dpi) / baseScreenDPI) * baseDPI;
     double scale = targetDPI / 96;
     if (scale < 1) {
         targetDPI = 96;
@@ -708,4 +764,61 @@ void XCBUI::setEnableTray(bool enable) {
         updateTray();
     }
 }
+
+void XCBUI::scheduleUpdateScreen() {
+    initScreenEvent_->setNextInterval(100000);
+    initScreenEvent_->setOneShot();
+}
+
+bool XCBUI::grabPointer(XCBWindow *window) {
+    auto cookie = xcb_grab_pointer(
+        conn_, false, window->wid(),
+        (XCB_EVENT_MASK_BUTTON_PRESS | XCB_EVENT_MASK_BUTTON_RELEASE |
+         XCB_EVENT_MASK_BUTTON_MOTION | XCB_EVENT_MASK_ENTER_WINDOW |
+         XCB_EVENT_MASK_LEAVE_WINDOW | XCB_EVENT_MASK_POINTER_MOTION),
+        XCB_GRAB_MODE_ASYNC, XCB_GRAB_MODE_ASYNC, XCB_WINDOW_NONE,
+        XCB_CURSOR_NONE, XCB_TIME_CURRENT_TIME);
+
+    auto reply = makeUniqueCPtr(xcb_grab_pointer_reply(conn_, cookie, nullptr));
+    if (reply && reply->status == XCB_GRAB_STATUS_SUCCESS) {
+        pointerGrabber_ = window;
+        return true;
+    }
+    return false;
+}
+
+void XCBUI::ungrabPointer() {
+    if (pointerGrabber_) {
+        xcb_ungrab_pointer(conn_, XCB_TIME_CURRENT_TIME);
+        pointerGrabber_ = nullptr;
+    }
+}
+
+void XCBUI::destroyCairoDevice(cairo_device_t *device) {
+    if (!device) {
+        return;
+    }
+    // Here's how cairo-xcb works.
+    // When a new xcb connection is seen, a internal shared xcb device is
+    // created and referenced by cairo itself. When device_finish is called,
+    // this internal shared reference is decreased, and will be destroyed when
+    // ref is 0. So while it looks like we reference and dereference which
+    // should does nothing, the cairo_device_finish will to the actual work to
+    // clean up the internal reference. See also:
+    // https://lists.cairographics.org/archives/cairo/2018-November/028791.html
+    // Such design is to allow xcb shared connection data to be kept even if
+    // there is no xcb surface. Though this API design really sucks, since there
+    // is no API to create a xcb device and pass it to cairo_xcb_surface_create.
+    // Instead, we have to get the device pointer from a xcb surface.
+    cairo_device_finish(device);
+    cairo_device_destroy(device);
+}
+
+void XCBUI::setCairoDevice(cairo_device_t *device) {
+    if (device_.get() != device) {
+        device_.reset();
+        device_.reset(cairo_device_reference(device));
+    }
+}
+
 } // namespace fcitx::classicui
