@@ -4,10 +4,18 @@
  * SPDX-License-Identifier: LGPL-2.1-or-later
  *
  */
+#include <memory>
+#include <string>
 #include "fcitx-utils/eventdispatcher.h"
+#include "fcitx-utils/handlertable.h"
+#include "fcitx-utils/key.h"
+#include "fcitx-utils/keysym.h"
+#include "fcitx-utils/log.h"
+#include "fcitx-utils/macros.h"
 #include "fcitx-utils/testing.h"
 #include "fcitx/addonmanager.h"
 #include "fcitx/instance.h"
+#include "fcitx/userinterface.h"
 #include "quickphrase_public.h"
 #include "testdir.h"
 #include "testfrontend_public.h"
@@ -16,8 +24,8 @@ using namespace fcitx;
 
 std::unique_ptr<HandlerTableEntry<QuickPhraseProviderCallback>> handle;
 
-void scheduleEvent(EventDispatcher *dispatcher, Instance *instance) {
-    dispatcher->schedule([instance]() {
+void testInit(Instance *instance) {
+    instance->eventDispatcher().schedule([instance]() {
         auto *quickphrase = instance->addonManager().addon("quickphrase", true);
         handle = quickphrase->call<IQuickPhrase::addProvider>(
             [](InputContext *, const std::string &text,
@@ -31,7 +39,10 @@ void scheduleEvent(EventDispatcher *dispatcher, Instance *instance) {
             });
         FCITX_ASSERT(quickphrase);
     });
-    dispatcher->schedule([dispatcher, instance]() {
+}
+
+void testBasic(Instance *instance) {
+    instance->eventDispatcher().schedule([instance]() {
         auto *testfrontend = instance->addonManager().addon("testfrontend");
         for (const auto *expectation :
              {"TEST", "abc", "abcd", "DEF", "abcd", "DEF1", "test1", "CALLBACK",
@@ -150,12 +161,36 @@ void scheduleEvent(EventDispatcher *dispatcher, Instance *instance) {
         testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("u"), false);
         testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("t"), false);
         testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("o"), false);
+    });
+}
 
-        dispatcher->schedule([dispatcher, instance]() {
-            handle.reset();
-            dispatcher->detach();
-            instance->exit();
-        });
+void testProviderV2(Instance *instance) {
+    instance->eventDispatcher().schedule([instance]() {
+        auto *testfrontend = instance->addonManager().addon("testfrontend");
+        testfrontend->call<ITestFrontend::pushCommitExpectation>("PROVIDERV2");
+        auto uuid =
+            testfrontend->call<ITestFrontend::createInputContext>("testapp");
+        testfrontend->call<ITestFrontend::keyEvent>(
+            uuid, Key(FcitxKey_BackSpace), false);
+        auto *quickphrase = instance->addonManager().addon("quickphrase", true);
+        auto handleV2 = quickphrase->call<IQuickPhrase::addProviderV2>(
+            [](InputContext *, const std::string &text,
+               const QuickPhraseAddCandidateCallbackV2 &callback) {
+                FCITX_INFO() << "Quickphrase text: " << text;
+                if (text == "PVD") {
+                    callback("PROVIDERV2", "V2", "COMMENT",
+                             QuickPhraseAction::Commit);
+                    return false;
+                }
+                return true;
+            });
+
+        testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("Super+grave"),
+                                                    false);
+        testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("P"), false);
+        testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("V"), false);
+        testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("D"), false);
+        testfrontend->call<ITestFrontend::keyEvent>(uuid, Key("1"), false);
     });
 }
 
@@ -172,9 +207,13 @@ int main() {
     char *argv[] = {arg0, arg1, arg2};
     Instance instance(FCITX_ARRAY_SIZE(argv), argv);
     instance.addonManager().registerDefaultLoader(nullptr);
-    EventDispatcher dispatcher;
-    dispatcher.attach(&instance.eventLoop());
-    scheduleEvent(&dispatcher, &instance);
+    testInit(&instance);
+    testBasic(&instance);
+    testProviderV2(&instance);
+    instance.eventDispatcher().schedule([&instance]() {
+        handle.reset();
+        instance.exit();
+    });
     instance.exec();
     return 0;
 }
